# Copyright 2016-2018 Carl Poelking, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import print_function
from __future__ import division
from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import str
from past.utils import old_div
from builtins import object
# See git https://github.com/capoe/momo.git
import os
import sys
import subprocess
import argparse
import time
import numpy as np
try:
    from lxml import etree
except ImportError:
    pass

boolean_dict = \
    {'true' : True,   '1' : True,  'yes' : True,
     'false' : False, '0' : False, 'no' : False, 'none' : False }

# =============================================================================
# XML WRAPPERS
# =============================================================================

class ExtendableNamespace(argparse.Namespace):
    def AddNamespace(self,  **kwargs):
        for name in kwargs:
            att = getattr(self, name, None)
            if att is None:
                setattr(self, name, kwargs[name])
            else:
                setattr(self, name, kwargs[name].As(type(att)))
        return
    def Add(self, name, value):
        att = getattr(self, name, None)
        if att is None:
            setattr(self, name, value)
        else:
            att.Add(name, value)
        return value

def GenerateTreeDict(tree, element, path='', paths_rel_to=None):
    if type(element) == etree._Comment: return [], {}
    # Update path
    if path == '':
        if element.tag != paths_rel_to:
            path += element.tag
    else:
        path += '/' + element.tag
    # Containers for lower levels
    tag_node = {}
    nodes = []
    # Construct Node
    xmlnode = XmlNode(element, path) # tree.getpath(element))
    nodes.append(xmlnode)
    if len(element) == 0:
        tag_node[path] = xmlnode
    #else:
    #	print "len 0", xmlnode.path
    #	tag_node[path] = xmlnode
    # Iterate over children
    for child in element:
        child_elements, childtag_element = GenerateTreeDict(tree, child, path)
        nodes = nodes + child_elements
        for key in childtag_element.keys():
            if key in tag_node:
                if type(tag_node[key]) != list:
                    tag_node[key] = [ tag_node[key], childtag_element[key] ]
                else:
                    tag_node[key].append(childtag_element[key])
            else:
                tag_node[key] = childtag_element[key]
    return nodes, tag_node

def NamespaceFromDict(tree_dict):
    nspace = ExtendableNamespace()
    for key in tree_dict.keys():
        #print "NSPACE Path = %s" % key
        sections = key.split('/')
        values = [ None for s in sections ]
        values[-1] = tree_dict[key]
        add_to_nspace = nspace
        for s,v in zip(sections, values):
            if v == None:
                #print "NSPACE Check for existing"
                if getattr(add_to_nspace, s, None):
                    #print "NSPACE '%s' already exists" % s
                    add_to_nspace = getattr(add_to_nspace, s, None)
                else:
                    #print "NSPACE '%s' created" % s
                    sub_nspace = ExtendableNamespace()
                    add_to_nspace = add_to_nspace.Add(s, sub_nspace)
            else:
                #print "NSPACE '%s' value added" % s
                add_to_nspace.Add(s, v)
    return nspace


class XmlTree(list):
    def __init__(self, xmlfile, paths_rel_to=None):
        self.xmlfile = xmlfile
        self.xtree = etree.parse(xmlfile)
        self.xroot = self.xtree.getroot()
        self.nodes, self.tag_node = GenerateTreeDict(self.xtree, self.xroot, '', paths_rel_to)
        self.xspace = NamespaceFromDict(self.tag_node)
    def SelectByTag(self, tag):
        selection = [ e for e in self.nodes if e.tag == tag ]
        return selection
    def __getitem__(self, key):
        return self.tag_node[key]
    def keys(self):
        return list(self.tag_node.keys())


class XmlNode(object):
    def __init__(self, element, path):
        self.path = path
        self.node = element
        self.tag = element.tag
        self.value = element.text
        self.attributes = element.attrib
    def As(self, typ):
        if typ == np.array:
            sps = self.value.split()
            return typ([ float(sp) for sp in sps ])
        elif typ == bool:
            return boolean_dict.get(self.value.lower())
        else:
            return typ(self.value)
    def AsArray(self, typ, sep=' ', rep='\t\n'):
        for r in rep:
            self.value = self.value.replace(r, sep)
        sp = self.value.split(sep)
        return [ typ(s) for s in sp if str(s) != '' ]
    def SetNodeValue(self, new_value):
        self.value = new_value
        if self.node != None:
            self.node.firstChild.nodeValue = new_value
        return
    def __getitem__(self, key):
        return self.node.get(key)


# =============================================================================
# COMMAND LINE & XML INPUT INTERFACE
# =============================================================================

class CLIO_HelpFormatter(argparse.HelpFormatter):
    def _format_usage(self, usage, action, group, prefix):
        return "%s : Command Line Interface\n" % sys.argv[0]


class OptionsInterface(object):
    def __init__(self):
        # COMMAND-LINE ARGUMENTS
        self.is_connected_to_cmd_ln = False
        self.cmd_ln_args = None
        self.cmd_ln_opts = None
        self.cmd_ln_nicknames = [ '-h' ]
        self.boolean_translator = boolean_dict
        self.subtype = str
        # XML OPTIONS FILE
        self.is_connected_to_xml = False
        self.xmlfile = None
        self.tree = None
        self.xdict = None
        self.xspace = None
        # JOINED OPTIONS
        self.opts = ExtendableNamespace()
    def Connect(self, xmlfile=None):
        self.ConnectToCmdLn()
        self.ConnectToOptionsFile(xmlfile)
    def Parse(self, xkey='options'):
        if self.is_connected_to_cmd_ln:
            self.ParseCmdLn()
        if self.is_connected_to_xml:
            self.ParseOptionsFileXml(xkey)
        if self.is_connected_to_cmd_ln and not self.is_connected_to_xml:
            return self.cmd_ln_opts
        elif self.is_connected_to_xml and not self.is_connected_to_cmd_ln:
            return self.xspace
        else:
            return self.cmd_ln_opts, self.xspace
    def ParseOptionsFile(self, xmlfile, xkey):
        self.xmlfile = xmlfile
        self.is_connected_to_xml = True
        self.ParseOptionsFileXml(xkey)
        return self.xspace
    # COMMAND-LINE CONVENIENCE ================================================
    def __call__(self):
        return self.cmd_ln_opts
    def ConnectToCmdLn(self, prog=sys.argv[0], descr=None):
        self.cmd_ln_args = argparse.ArgumentParser(prog=sys.argv[0],
            formatter_class=lambda prog: CLIO_HelpFormatter(prog,max_help_position=70))
        self.is_connected_to_cmd_ln = True
        return
    def ParseCmdLn(self):
        self.cmd_ln_opts = self.cmd_ln_args.parse_args()
    def InterpretAsBoolean(self, expr):
        try:
            return self.boolean_translator.get(expr.lower())
        except KeyError:
            raise ValueError('CLIO does not know how to convert %s into a boolean.' % expr)
    def InterpretAsNumpyArray(self, expr):
        print("Interpret", expr)
        array = [ float(e) for e in expr ]
        array = np.array(array)
        return array
    def InterpretAsList(self, expr):
        array = [ self.subtype(e) for e in expr ]
        return array
    def AddArg(self, name, typ=str, nickname=None,
            default=None, destination=None, help=None):
        # Sort out <name> (e.g. --time) vs <destination> (e.g., time)
        if '--' != name[0:2]:
            dest = name
            name = '--' + name
        else:
            dest = name[2:]
        # Sort out <default> vs <required>
        if default == None: required = True
        else: required = False
        # Construct default <help> it not given
        if help == None: help = str(typ)
        # Construct <nickname> if not given
        if nickname == None:
            nickname = '-'
            for char in dest:
                nickname += char
                if not nickname in self.cmd_ln_nicknames:
                    break
            if nickname in self.cmd_ln_nicknames:
                raise ValueError('CLIO could not construct nickname from %s option'\
                    % name)
            self.cmd_ln_nicknames.append(nickname)
        # Process type
        if typ in [int, float, str]:
            nargs = None
        elif typ == bool:
            typ = self.InterpretAsBoolean
            nargs = None
        elif typ == np.array:
            raise NotImplementedError
            typ = float # self.InterpretAsNumpyArray
            nargs = 3
        elif typ == list:
            typ = str
            nargs = '*'
        elif len(typ) == 2 and typ[0] == list:
            typ = typ[1]
            nargs = '*'
        else:
            raise NotImplementedError("CLIO does not know how to generate type '%s'"\
                 % typ)
        self.cmd_ln_args.add_argument(nickname, name,
                dest=dest,
                action='store',
                nargs=nargs,
                required=required,
                type=typ,
                metavar=dest[0:1].upper(),
                default=default,
                help=help)
        return
    # COMMAND-LINE CONVENIENCE ================================================
    def ConnectToOptionsFile(self, xmlfile):
        if xmlfile == None or xmlfile == '': return
        self.xmlfile = xmlfile
        self.is_connected_to_xml = True
        return
    def ParseOptionsFileXml(self, xkey='options'):
        if self.xmlfile == None: return
        self.tree = XmlTree(self.xmlfile, paths_rel_to=xkey)
        self.xdict = self.tree.tag_node
        self.xspace = self.tree.xspace
        return
    def __getitem__(self, key):
        try:
            return self.xspace.__dict__[key]
        except KeyError:
            return self.cmd_ln_opts.__dict__[key]
        except KeyError:
            raise AttributeError('No such option registered: \'%s\'' % key)
        return None


# =============================================================================
# OS SHELL INTERFACE
# =============================================================================

class ShellInterface(object):
    def __init__(self):
        # =====================================================================
        # PRINTER ATTRIBUTES
        self.color_dict = { \
            'pp' : '\033[95m',
            'mb' : '\033[34m',
            'lb' : '\033[1;34m',
            'my' : '\033[1;33m',
            'mg' : '\033[92m',
            'mr' : '\033[91m',
            'ww' : '\033[0;1m',
            'ok' : '\033[92m',
            'xx' : '\033[91m',
            'warning' : '\033[93m',
            'error' : '\033[95m',
            'endcolor' : '\033[0;1m' }
        self.justify_dict = { \
            'o' : '  o ',
            '.' : '... ',
            'r' : '\r',
            'ro' : '\r  o '}
        self.pp = OS_COLOR('pp')
        self.lb = OS_COLOR('lb')
        self.mb = OS_COLOR('mb')
        self.mg = OS_COLOR('mg')
        self.my = OS_COLOR('my')
        self.mr = OS_COLOR('mr')
        self.ww = OS_COLOR('ww')
        self.ok = OS_COLOR('ok')
        self.xx = OS_COLOR('xx')
        self.colors = [ OS_COLOR(c) for c in sorted(self.color_dict.keys()) ]
        self.item = '  o '
        self.iitem = '      - '
        self.endl = OS_LINE_CHAR('\n')
        self.flush = OS_LINE_CHAR('')
        self.back = OS_LINE_CHAR('\r')
        self.trail = ' '
        # CURRENT STYLE SELECTION
        self.sel_color = None
        self.sel_justify = None
        self.sel_header = False
        self.sel_trim = "="
        # EXE ATTRIBUTES
        self.catch = OS_EXE_CATCH()
        self.assert_zero = OS_EXE_ASSERT()
        self.dev = OS_EXE_DEV('')
        self.nodev = OS_EXE_DEV('')
        self.devnull = OS_EXE_DEV()
        self.devfile = OS_EXE_DEV('')
        self.os_exe_get = False
        self.os_exe_assert_zero = False
        self.os_exe_dev = ''
        self.os_exe_verbose = False
        # LOGGING
        self.logfile = None
        # DIRECTORY HOPPING
        self.paths_visited = [ os.getcwd() ]
        self.exe_root_path = self.paths_visited[0]
        self.N_store_paths_visited = 1+5
    def __call__(self, mssg, c=None, j=None, h=False, t="="):
        # c=color, j=justify, h=header, t=trim, u=upper-case
        if j:
            mssg = self.justify_dict[j] + mssg
        if c != None:
            mssg = self.color_dict[c] + mssg + self.color_dict['endcolor']
        if h:
            mssg = self.os_generate_header(mssg, t)
        print(mssg)
    # LOGFILE ADAPTOR =========================================================
    def ConnectToFile(self, logfile):
        self.logfile = logfile
        sys.stdout = open(logfile, 'w')
        self.devfile = OS_EXE_DEV(' >> {log} 2>> {log}'.format(log=logfile))
        return
    def DisconnectFromFile(self):
        if self.logfile != None:
            self.devfile = OS_EXE_DEV('')
            self.logfile = None
            sys.stdout = sys.__stdout__
        else: pass
        return
    # PRINTER METHODS =========================================================
    def __lshift__(self, mssg):
        if type(mssg) == OS_LINE_CHAR:
            # <LOG MESSAGE HERE>
            sys.stdout.write(str(mssg))
            sys.stdout.flush()
            self.sel_color = None
            return self
        elif type(mssg) == OS_COLOR:
            self.sel_color = str(mssg)
            return self
        mssg = str(mssg)
        if self.sel_justify != None:
            mssg = self.justify_dict[j] + mssg
        mssg += self.trail
        if self.sel_color != None:
            mssg = self.color_dict[self.sel_color] \
                + mssg + self.color_dict['endcolor']
        if self.sel_header:
            mssg = self.os_generate_header(mssg, self.sel_trim)
        # <LOG MESSAGE HERE>
        sys.stdout.write(mssg)
        return self
    def os_print(self, mssg, c=None, j=None, h=False, t="="):
        # c=color, j=justify, h=header, t=trim, u=upper-case
        if j:
            mssg = OS_JUSTIFY_DICT[j] + mssg
        if c != None:
            mssg = OS_COLOR_DICT[c] + mssg + OS_COLOR_DICT['endcolor']
        if h:
            mssg = os_generate_header(mssg, t)
        print(mssg)
        return
    def os_print_config(self, c=None, j=None, h=False, t="=", tl=' '):
        self.sel_color = c
        self.sel_justify = j
        self.sel_header = h
        self.sel_trim = t
        self.trail = tl
        return
    def os_print_reset(self):
        self.sel_color = None
        self.sel_justify = None
        self.sel_header = False
        self.sel_trim = "="
        self.trail = ' '
        return
    def os_generate_header(title, trim="="):
        try:
            height, width = os.popen('stty size', 'r').read().split()
            width = int(width)
            leftright = (width - len(title)-2) // 2
        except ValueError:
            leftright = 40
        return trim*leftright + " " + title + " " + trim*leftright
    # EXECUTE METHODS =========================================================
    def __rshift__(self, cmmd):
        if type(cmmd) == OS_EXE_CATCH:
            self.os_exe_get = True
            return self
        elif type(cmmd) == OS_EXE_DEV:
            self.dev = cmmd
            return self
        elif type(cmmd) == OS_EXE_ASSERT:
            self.os_exe_assert_zero = True
            return self
        # Redirect command as requested (not applicable if output is asked for)
        if not self.os_exe_get:
            if str(self.dev) != '':
                cmmd += str(self.dev)
                self.dev = self.nodev
            else:
                cmmd += str(self.devfile)
        # Execute
        if self.debug: self << self.my << "exe:" << cmmd << endl
        if self.os_exe_get:
            output = subprocess.getoutput(cmmd)
            self.os_exe_get = False
            return output
        else:
            sign = os.system(cmmd)
        if self.os_exe_assert_zero:
            if str(sign) != '0':
                raise RuntimeError("<OSIO> '%s' returned '%s'" % (cmmd, sign))
            self.os_exe_assert_zero = False
        return sign
    # EXIT PROTOCOLS ==========================================================
    def okquit(self, what=''):
        if what != '': self << self.ok << what << self.endl
        self.DisconnectFromFile()
        sys.exit(0)
        return
    def xxquit(self, what=''):
        if what != '': self << self.xx << "ERROR" << what << self.endl
        self.DisconnectFromFile()
        sys.exit(1)
        return
    # DIRECTORY HOPPING =======================================================
    def cd(self, d):
        # Current working directory, for archiving ... =>
        cwd = os.getcwd()
        if type(d) == int:
            # Change to previously visited path
            os.chdir(self.paths_visited[d])
        elif type(d) == str:
            # Change to path as specified explicitly
            os.chdir(d)
        else:
            raise NotImplementedError
        # <= ... previous path
        self.paths_visited.append(cwd)
        if len(self.paths_visited) > self.N_store_paths_visited:
            self.paths_visited.pop(1) # 0 stores root
        if self.debug: self << self.my << "cd: " << os.getcwd() << self.endl
        return
    def pwd(self):
        return self.cwd()
    def cwd(self):
        return os.getcwd()
    def root(self):
        self.cd(self.exe_root_path)
        return
    def abspath(self, file):
        if not os.path.exists(file):
            raise IOError("<osio::abspath> No such item in local directory: '%s'" % file)
        return os.path.join(self.cwd(), file)
    def mkcd(self, directory):
        self >> self.assert_zero >> 'mkdir -p %s' % directory
        self.cd(directory)
        return directory

class OS_EXE_DEV(object):
    def __init__(self, dev=' > /dev/null 2> /dev/null'):
        self.dev = dev
    def __str__(self):
        return self.dev

class OS_EXE_CATCH(object):
    def __init__(self):
        self.catch = True

class OS_EXE_ASSERT(object):
    def __init__(self):
        self.assert_0 = True

class OS_COLOR(object):
    def __init__(self, colstr):
        self.colstr = colstr
    def __str__(self):
        return self.colstr

class OS_LINE_CHAR(object):
    def __init__(self, char):
        self.char = char
    def __str__(self):
        return self.char


class OSIO(ShellInterface, OptionsInterface):
    def __init__(self):
        self.debug = False
        ShellInterface.__init__(self)
        OptionsInterface.__init__(self)
        return
    def sleep(self, dt):
        time.sleep(dt)
        return


osio = OSIO()
endl = OS_LINE_CHAR('\n')
flush = OS_LINE_CHAR('')
back = OS_LINE_CHAR('\r')
catch = OS_EXE_CATCH()
devnull = OS_EXE_DEV()




# Copyright 2016-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Used to read the CPMD manual .tex file and generate the input file structure
from it. From the structure one can automatically generate the metainfo
definitions and the pickle file that is used to validate the structure during
parsing.
"""
import re
import json
import pickle
from cpmdparser.generic.inputparsing import Section, Keyword, metainfo_data_prefix, metainfo_section_prefix


#===============================================================================
def generate_input_tree(filepath):

    with open(filepath, "r") as fin:
        lines = fin.readlines()

    # Create the sections
    section_regex_string = "\\\> \\\&(\w+) \.\.\.\s+\\\> \\\&END  \\\>  \$"
    section_regex = re.compile(section_regex_string)
    root_section = Section("", "The root section for CPMD input. Contains all the other input sections.")

    for i_line in range(len(lines)):
        line = lines[i_line]
        match = section_regex.match(line)
        if match:
            section_name = match.groups()[0]
            section = Section(section_name)

            desc_end = "\>"
            desc_ended = False
            description = []
            i_line_desc = 1
            while not desc_ended:
                desc_line = lines[i_line + i_line_desc]
                i_line_desc += 1
                if desc_line.startswith(desc_end):
                    desc_ended = True
                else:
                    description.append(desc_line.strip()[0:-2])
            section.description = " ".join(description)
            if section_name in root_section.subsections:
                raise KeyError("Multiple sections with the same name")
            root_section.subsections[section_name] = section

    # Create the keywords
    keyword_regex_string = r"\\keyword\{([\w ]+)\}\{(.*?)\}\{(.*?)\}\{(.*?)\}\{\\&(\w+)\}"
    keyword_regex = re.compile(keyword_regex_string)
    spekeyword_regex_string = r"\\spekeyword\{(\w+)\}\{(.*?)\}\{(.*?)\}\{(.*?)\}\{\\&(\w+)\}\{(\w+)\}"
    spekeyword_regex = re.compile(spekeyword_regex_string)

    for i_line in range(len(lines)):
        line = lines[i_line]
        keywordmatch = keyword_regex.match(line)
        spekeywordmatch = spekeyword_regex.match(line)
        if keywordmatch or spekeywordmatch:
            if keywordmatch:
                groups = keywordmatch.groups()
            elif spekeywordmatch:
                groups = spekeywordmatch.groups()
            keyword_name = groups[0]
            # print(keyword_name)
            # if keyword_name == "OPTIMIZE WAVEFUNCTION":
                # print("FOUND")
            available_options1 = groups[1]
            available_options2 = groups[2]
            parent_section_name = groups[4]
            keyword = Keyword(keyword_name)
            if spekeywordmatch:
                unique_name = groups[5]
                keyword.unique_name = unique_name
            else:
                keyword.unique_name = keyword_name

            # Parse the available options in the first list
            if available_options1:
                available_options = available_options1.split(",")
                for option in available_options:
                    if "=" in option:
                        option_split = option.split("=")
                        option_name = option_split[0]
                    else:
                        option_name = option
                    option_name = re.sub(r"[\{\}\[\]\\ ]", "", option_name)
                    keyword.available_options.append(option_name)
                # print(keyword.available_options)

            # Parse the available options in the second list
            if available_options2:
                available_options = available_options2.split(",")
                for option in available_options:
                    if "=" in option:
                        option_split = option.split("=")
                        option_name = option_split[0]
                    else:
                        option_name = option
                    option_name = re.sub(r"[\{\}\[\]\\ ]", "", option_name)
                    keyword.available_options.append(option_name)
                # print(keyword.available_options)

            # Parse the description
            desc_ended = False
            description = []
            i_line_desc = 1
            n_braces = 0
            while not desc_ended:
                desc_line = lines[i_line + i_line_desc]
                i_line_desc += 1
                for i_char, character in enumerate(desc_line):
                    if character == "{":
                        n_braces += 1
                    elif character == "}":
                        n_braces -= 1
                        if n_braces == 0:
                            # if desc_line.startswith("\desc{"):
                                # desc_line = desc_line
                            description.append(desc_line[0:i_char].strip())
                            desc_ended = True
                if n_braces != 0:
                    full_line = desc_line.strip()
                    # if full_line.startswith("\desc{"):
                        # full_line = full_line[6:]
                    if full_line.endswith(r"\\"):
                        full_line = full_line[:-2]
                    if full_line == "%":
                        full_line = "\n"
                    description.append(full_line)
            keyword.description = " ".join(description)
            if keyword.description.startswith("\desc{"):
                keyword.description = keyword.description[6:]
                # print("DESC")

            parent_section = root_section.subsections[parent_section_name]
            parent_section.keywords[keyword_name].append(keyword)

    return root_section


#===============================================================================
def generate_pickle(filepath):
    input_tree = generate_input_tree(filepath)
    file_name = "../versions/cpmd41/input_data/cpmd_input_tree.pickle"
    fh = open(file_name, "wb")
    pickle.dump(input_tree, fh, protocol=2)


#===============================================================================
def generate_input_metainfos(filepath):

    json_root = {
        "type": "nomad_meta_info_1_0",
        "description": "Metainfo for the values parsed from a CPMD input file.",
        "dependencies": [ {
            "relativePath": "cpmd.general.nomadmetainfo.json"
            }],
    }

    root = generate_input_tree(filepath)
    parent = None
    root.name = ""
    root.description = "Contains the CPMD input file contents."
    container = []
    generate_metainfo_recursively(root, parent, container)
    json_root["metaInfos"] = container
    with open("input_metainfo.json", "w") as f:
        f.write(json.dumps(json_root, indent=2, separators=(',', ': ')))


#===============================================================================
def generate_metainfo_recursively(section, parent, container):

    json = generate_section_metainfo_json(section, parent)
    container.append(json)
    for subsection in section.subsections.values():
        generate_metainfo_recursively(subsection, section, container)
    for keyword_list in section.keywords.values():
        for keyword in keyword_list:
            key_json = generate_keyword_metainfo_json(keyword, section)
            options_json = generate_options_metainfo_json(keyword, section)
            parameter_json = generate_parameter_metainfo_json(keyword, section)
            container.append(key_json)
            container.append(options_json)
            container.append(parameter_json)
    if parent is not None:
        def_json = generate_default_keyword_metainfo_json(section)
    container.append(def_json)


#===============================================================================
def generate_keyword_metainfo_json(keyword, section):
    json_obj = {}
    json_obj["name"] = metainfo_section_prefix + "{}.{}".format(section.name, keyword.unique_name.replace(" ", "_"))
    json_obj["superNames"] = [metainfo_section_prefix + "{}".format(section.name)]

    # Description
    description = keyword.description
    if description is None or description.isspace():
        description = "Settings for {}".format(keyword.unique_name.replace(" ", "_"))
    json_obj["description"] = description
    json_obj["kindStr"] = "type_section"

    return json_obj


#===============================================================================
def generate_parameter_metainfo_json(keyword, section):
    json_obj = {}
    json_obj["name"] = metainfo_data_prefix + "{}.{}_parameters".format(section.name, keyword.unique_name).replace(" ", "_")
    json_obj["description"] = "The parameters for keyword {}.".format(keyword.unique_name.replace(" ", "_"))
    json_obj["superNames"] = [metainfo_section_prefix + "{}.{}".format(section.name, keyword.unique_name.replace(" ", "_"))]
    json_obj["dtypeStr"] = "C"
    json_obj["shape"] = []

    return json_obj


#===============================================================================
def generate_options_metainfo_json(keyword, section):
    json_obj = {}
    json_obj["name"] = metainfo_data_prefix + "{}.{}_options".format(section.name, keyword.unique_name.replace(" ", "_"))
    json_obj["description"] = "The options given for keyword {}.".format(keyword.unique_name.replace(" ", "_"))
    json_obj["superNames"] = [metainfo_section_prefix + "{}.{}".format(section.name, keyword.unique_name.replace(" ", "_"))]
    json_obj["dtypeStr"] = "C"
    json_obj["shape"] = []

    return json_obj


#===============================================================================
def generate_default_keyword_metainfo_json(section):
    json_obj = {}
    json_obj["name"] = metainfo_data_prefix + "{}_default_keyword".format(section.name)
    json_obj["description"] = "The parameters that are present in the section {} even without a keyword.".format(section.name)
    json_obj["superNames"] = [metainfo_section_prefix + "{}".format(section.name)]
    json_obj["dtypeStr"] = "C"
    json_obj["shape"] = []

    return json_obj


#===============================================================================
def generate_section_metainfo_json(section, parent):
    json_obj = {}

    if parent is None:
        json_obj["name"] = "x_cpmd_section_input"
        json_obj["superNames"] = ["section_run"]
    else:
        json_obj["name"] = metainfo_section_prefix + "{}".format(section.name.replace(" ", "_"))
        json_obj["superNames"] = ["x_cpmd_section_input"]
    json_obj["kindStr"] = "type_section"

    description = section.description
    if description is None or description.isspace():
        description = "Settings for {}".format(section.name)
    json_obj["description"] = description

    return json_obj


#===============================================================================
if __name__ == "__main__":
    filepath = "../versions/cpmd41/input_data/manual.tex"
    generate_pickle(filepath)
    # generate_input_metainfos(filepath)



from optimade.models import StructureResource as OptimadeStructure

from optimade.adapters.structures.utils import pad_cell, pad_positions

try:
    from aiida.orm.nodes.data.structure import StructureData, Kind, Site
except (ImportError, ModuleNotFoundError):
    from warnings import warn

    StructureData = None
    AIIDA_NOT_FOUND = (
        "AiiDA not found, cannot convert structure to an AiiDA StructureData"
    )


__all__ = ("get_aiida_structure_data",)


def get_aiida_structure_data(optimade_structure: OptimadeStructure) -> StructureData:
    """ Get AiiDA StructureData from OPTIMADE structure
    :param optimade_structure: OPTIMADE structure
    :return: StructureData
    """
    if globals().get("StructureData", None) is None:
        warn(AIIDA_NOT_FOUND)
        return None

    attributes = optimade_structure.attributes

    # Convert null/None values to float("nan")
    lattice_vectors, adjust_cell = pad_cell(attributes.lattice_vectors)
    structure = StructureData(cell=lattice_vectors)

    # Add Kinds
    for kind in attributes.species:
        symbols = []
        concentration = []
        for index, chemical_symbol in enumerate(kind.chemical_symbols):
            # NOTE: The non-chemical element identifier "X" is identical to how AiiDA handles this,
            # so it will be treated the same as any other true chemical identifier.
            if chemical_symbol == "vacancy":
                # Skip. This is how AiiDA handles vacancies;
                # to not include them, while keeping the concentration in a site less than 1.
                continue
            else:
                symbols.append(chemical_symbol)
                concentration.append(kind.concentration[index])

        # AiiDA needs a definition for the mass, and for it to be > 0
        # mass is OPTIONAL for OPTIMADE structures
        mass = kind.mass if kind.mass else 1

        structure.append_kind(
            Kind(symbols=symbols, weights=concentration, mass=mass, name=kind.name)
        )

    # Convert null/None values to float("nan")
    cartesian_site_positions, _ = pad_positions(attributes.cartesian_site_positions)

    # Add Sites
    for index in range(attributes.nsites):
        # range() to ensure 1-to-1 between kind and site
        structure.append_site(
            Site(
                kind_name=attributes.species_at_sites[index],
                position=cartesian_site_positions[index],
            )
        )

    if adjust_cell:
        structure._adjust_default_cell(
            pbc=[bool(dim.value) for dim in attributes.dimension_types]
        )

    return structure



#!/usr/bin/env python
from __future__ import print_function
from past.builtins import cmp
from builtins import range
from builtins import object
import sys, os, os.path, datetime, logging, json
basePath = os.path.realpath(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
if not basePath in sys.path:
    sys.path.append(basePath)
from nomadcore.local_meta_info import InfoKindEnv, InfoKindEl, loadJsonFile, loadJsonStream
from nomadcore.json_support import jsonIndentF, jsonCompactF
import git

class GitTreeDependencySolver(object):
    def __init__(self, tree, basePath, context):
        self.tree = tree
        self.basePath = basePath
        self.context = context
        self.deps = {}

    def __call__(self, infoKindEnv, source, dep):
        if "relativePath" not in dep:
            raise Exception('Invalid dependency for relativeDependencySolver there must be a relativePath')
        dPath = os.path.normpath(os.path.join(self.basePath, dep.get('relativePath')))
        if dPath in self.deps:
            return self.deps[dPath]
        depInfo = None
        depIKEnv = InfoKindEnv(name = os.path.basename(dPath), path = dPath, dependencyLoader=infoKindEnv.dependencyLoader)
        self.deps[dPath] = depIKEnv
        try:
            f=self.tree[dPath].data_stream
        except:
            print(self.tree.hexsha, ", t.path", self.tree.path, ", bpath:", self.basePath, ", dPath:", dPath)
            raise Exception("Failed to resolve dependency {0!r} in {context}, due to exception {1}".format(dep, sys.exc_info()[1], context = self.context))
        depInfo = json.load(f)
        if depInfo:
            depIKEnv.fromJsonList(depInfo, source = { 'path': dPath }, dependencyLoad = True)
        return depIKEnv

def loadPath(path, loadOverrides = False):
    """loads all nomadmetainfo.json files within the given path.
    If loadOverrides is true then also nomadmetainfooverrides.json files are loaded"""
    allEnvs = {}
    allOverrides = {}
    hasWarnings = False
    if os.path.isdir(path):
        for dirpath, dirnames, filenames in os.walk(path):
            for filename in filenames:
                if filename.endswith(".nomadmetainfo.json"):
                    filepath = os.path.join(dirpath, filename)
                    env, warnings = loadJsonFile(filepath, extraArgsHandling = InfoKindEl.ADD_EXTRA_ARGS)
                    if warnings.get("duplicates", False) or warnings.get("overrides", False):
                        hasWarnings = True
                        logging.warn("{0!r}: warnings in loadJsonFile:{1}\n".format(filepath, warnings))
                    allEnvs[os.path.relpath(filepath, path)] = env
                elif loadOverrides and filename.endswith(".nomadmetainfooverrides.json"):
                    try:
                        with open(filename) as f:
                            overrides = json.load(f)
                    except:
                        hasWarnings = True
                        logging.exception("Error loading %r", filename)
                    else:
                        allOverrides[os.path.relpath(filepath, path)] =  overrides
    return {"hasWarnings": hasWarnings, "envs": allEnvs, "overrides": allOverrides}

def loadRef(rev, path, loadOverrides = False):
    """loads all nomadmetainfo.json files within path from repo reference"""
    allEnvs = {}
    allOverrides = {}
    hasWarnings = False
    if path == ".":
        path = ""
    if path:
        tree = rev.tree[path]
    else:
        tree = rev.tree
    def loadBlobHasWarnings(obj):
        hasW = False
        if obj.type != 'blob':
            return False
        rPath = os.path.relpath(obj.path, path)
        if obj.name.endswith(".nomadmetainfo.json"):
            env, warnings = loadJsonStream(obj.data_stream, GitTreeDependencySolver(rev.tree, os.path.dirname(obj.path), obj.path),
                                           filePath = obj.path)
            allEnvs[rPath] = env
            dup = warnings.get("duplicates")
            hid = warnings.get("hidden")
            if dup or hid:
                hasW = True
                logging.warn("loading {path} of revision {rev}: {warnings}\n".format(
                    path = obj.path, rev = rev.name_rev, warnings = warnings))
        elif obj.name.endswith(".nomadmetainfooverrides.json"):
            try:
                overrides = json.load(obj.data_stream)
            except:
                hasW = True
                logging.exception("Error loading %r in revision %s", obj.path, rev.hexsha[:10])
            else:
                allOverrides[rPath] = overrides
        return hasW
    if tree.type == 'blob':
        if loadBlobHasWarnings(tree):
            hasWarnings = True
    else:
        for obj in tree.traverse(lambda i, d: i.type == 'blob' and (i.path.endswith(".nomadmetainfo.json") or
                                                                    (loadOverrides and i.path.endswith(".nomadmetainfooverrides.json")))):
            if loadBlobHasWarnings(obj):
                hasWarnings = True
    return {"hasWarnings": hasWarnings, "envs": allEnvs, "overrides": allOverrides}

def insertDicList(dict, key, value):
    if key in dict:
        dict[key].append(value)
    else:
        dict[key] = [value]

def prepareValues(values, name):
    "prepares the values computing gids"
    allNames = {}
    hasWarnings = False
    for path, env in values.items():
        try:
            env.calcGids()
        except:
            hasWarnings = True
            logging.warn("Error calculating gids of {path!r} {name}: {exc}\n".format(path = path, name = name, exc = sys.exc_info()[1]))
        noDepNames = env.noDepNames()
        for name, gid in env.gids.items():
            if name in noDepNames:
                insertDicList(allNames, name, (path, gid))
    return {"allNames": allNames, "envs": values, "hasWarnings": hasWarnings}

def cmpOverrides(o1,o2):
    c = cmp(o1["name"], o2["name"])
    if c != 0:
        return c
    return cmp(o1, o2)

def buildOverrides(oldValues, newValues):
    oldV = prepareValues(oldValues, "(old)")
    newV = prepareValues(newValues, "(new)")
    oldNames = oldV["allNames"]
    newNames = newV["allNames"]
    overrides = []
    complexOverrides = []
    additions = []
    removals = []
    hasWarnings = oldV["hasWarnings"] or newV["hasWarnings"]
    nMatched = 0
    for name, oldUses in oldNames.items():
        if not name in newNames:
            removals.append({"name": name, "oldGids": [x[1] for x in oldUses], "newGids": []})
            continue
        newUses = newNames[name]
        newNoMatch = list(newUses)
        oldNoMatch = []
        for oldPath, oldGid in oldUses:
            found = -1
            for i in range(len(newNoMatch)):
                newPath, newGid = newNoMatch[i]
                if newGid == oldGid:
                    del newNoMatch[i]
                    found = i
                    break
            if found != -1:
                nMatched += 1
            else:
                oldNoMatch.append((oldPath, oldGid))
        if not oldNoMatch and not newNoMatch:
            continue
        if len(oldNoMatch) == 1 and len(newNoMatch) == 1:
            overrides.append({"name": name, "oldGid": oldNoMatch[0][1], "newGid": newNoMatch[0][1]})
            continue
        # try path matching
        iOld = 0
        while iOld < len(oldNoMatch):
            oldPath, oldGid = oldNoMatch[iOld]
            found = -1
            for iNew in range(len(newNoMatch)):
                newPath, newGid = newNoMatch[iNew]
                if newPath == oldPath:
                    overrides.append({"name": name, "oldGid": oldGid, "newGid": newGid})
                    del newNoMatch[iNew]
                    del old
                    found = iNew
                    break
            if found != -1:
                del oldNoMatch[iOld]
            else:
                iOld += 1
        if len(oldNoMatch) == 1 and len(newNoMatch) == 1:
            overrides.append({"name": name, "oldGid": oldNoMatch[0][1], "newGid": newNoMatch[0][1]})
        if not oldNoMatch and not newNoMatch:
            continue
        elif oldNoMatch and not newNoMatch:
            removals.append({"name": name, "oldGids": [x[1] for x in oldNoMatch], "newGids": []})
        elif not oldNoMatch and newNoMatch:
            additions.append({"name": name, "oldGids": [], "newGids": [x[1] for x in newNoMatch]})
        else: # oldNoMatch and newNoMatch
            complexOverrides.append({"name": name, "oldGids": oldNoMatch, "newGids": newNoMatch[0][1]})
    for name in newNames.keys():
        if not name in oldNames:
            additions.append({"name": name, "oldGids": [], "newGids": [x[1] for x in newNames[name]]})
    overrides.sort(cmpOverrides)
    complexOverrides.sort(cmpOverrides)
    additions.sort(cmpOverrides)
    removals.sort(cmpOverrides)
    return {
        "oldNames": oldNames,
        "newNames": newNames,
        "overrides": overrides,
        "complexOverrides": complexOverrides,
        "additions": additions,
        "removals": removals,
        "hasWarnings": hasWarnings,
        "nMatched": nMatched
    }



if __name__ == "__main__":
    overridesDir = os.path.normpath(os.path.join(basePath,"../../../nomad-meta-info/meta_info/nomad_meta_info_overrides"))
    defaultPath = os.path.normpath(os.path.join(basePath,"../../../nomad-meta-info/meta_info/nomad_meta_info"))
    usage = """usage: {command} [--check-only] [--old-ref <ref1=HEAD>] [--old-path <path=None>] [--help]
    [--new-ref <ref2=None>] [--new-path <path=basePath>] [--repo-path <repoPath=None>]
    [--overrides-dir <overridesDir>] [--no-clobber] [--verbose] [<basePath>]

Generates the overrides for the kind info with the same name but different sha from the old version
to the new version.
Old version can be given as a revision in git (defaults to HEAD) with --old-ref,  or a path with
--old-path, likewise the new version can be given as a revision with --new-ref or new-path.
basePath (which defaults to {defaultPath}) gives the path of which the revisions should be compared (might be a
subdirectory or file of the repo) and if no --new-ref is given the version checked out there is
directly the new vesion.
The repository path by default is found looking from path upward for a .git repository, but can
also be given explicitly with --repo-path.
overridesDir defaults to {overridesDir}
By default output goes to stdout, but an explicit --out-file can also be given
""".format(command = os.path.basename(sys.argv[0]), overridesDir = overridesDir, defaultPath = defaultPath)
    oldRef = None
    oldPath = None
    newRef = None
    newPath = None
    repoPath = None
    path = None
    checkOnly = False
    noClobber = False
    verbose = False

    iArg = 1
    while iArg < len(sys.argv):
        arg = sys.argv[iArg]
        if arg == "--help":
            sys.stderr.write(usage)
            sys.exit(0)
        elif arg == "--check-only":
            checkOnly = True
        elif arg == "--no-clobber":
            noClobber = True
        elif arg == "--verbose":
            verbose = True
        elif arg == "--old-ref":
            iArg += 1
            if iArg < len(sys.argv):
                oldRef = sys.argv[iArg]
            else:
                sys.stderr.write("Error: missing reference after --old-ref\n\n")
                sys.stderr.write(usage)
                sys.exit(1)
        elif arg == "--old-path":
            iArg += 1
            if iArg < len(sys.argv):
                oldPath = sys.argv[iArg]
            else:
                sys.stderr.write("Error: missing path after --old-path\n\n")
                sys.stderr.write(usage)
                sys.exit(2)
        elif arg == "--new-ref":
            iArg += 1
            if iArg < len(sys.argv):
                newRef = sys.argv[iArg]
            else:
                sys.stderr.write("Error: missing reference after --new-ref\n\n")
                sys.stderr.write(usage)
                sys.exit(3)
        elif arg == "--new-path":
            iArg += 1
            if iArg < len(sys.argv):
                newPath = sys.argv[iArg]
            else:
                sys.stderr.write("Error: missing path after --new-path\n\n")
                sys.stderr.write(usage)
                sys.exit(4)
        elif arg == "--repo-path":
            iArg += 1
            if iArg < len(sys.argv):
                repoPath = sys.argv[iArg]
            else:
                sys.stderr.write("Error: missing path after --repo-path\n\n")
                sys.stderr.write(usage)
                sys.exit(5)
        elif arg == "--overrides-dir":
            iArg += 1
            if iArg < len(sys.argv):
                overridesDir = sys.argv[iArg]
            else:
                sys.stderr.write("Error: missing reference after --overrides-dir\n\n")
                sys.stderr.write(usage)
                sys.exit(7)
        elif path is None:
            path = arg
        else:
            sys.stderr.write("Error: unexpected extra argument {0!r}\n\n".format(arg))
            sys.stderr.write(usage)
            sys.exit(8)
        iArg += 1
    if path is None:
        path = defaultPath
    requiresRepo = True
    if oldPath is None and oldRef is None:
        oldRef = "HEAD"
    if not newRef and not newPath:
        newPath = path
    if oldPath and not newRef:
        requiresRepo = False
        if oldRef:
            sys.stderr.write("Error: both --old-path and --old-ref were given this is not supported\n\n")
            sys.stderr.write(usage)
            sys.exit(9)
    if requiresRepo and repoPath is None:
        repoPath = os.path.normpath(os.path.abspath(path))
        while not os.path.exists(os.path.join(repoPath, ".git")):
            if repoPath == '/':
                sys.stderr.write("Error: path {0!r} from is not within a git repo.\nCould not find .git directory, for bare repositories use the --repo-path option\n\n".format(arg))
                sys.stderr.write(usage)
                sys.exit(10)
            repoPath = os.path.split(repoPath)[0]
    if repoPath:
        repo = git.Repo(repoPath)
        if os.path.abspath(path).startswith(repoPath) and os.path.exists(os.path.abspath(path)):
            relPath = os.path.relpath(os.path.normpath(os.path.abspath(path)), repoPath)
        else:
            relPath = path
        if requiresRepo and relPath.startswith("../"):
            sys.stderr.write("Error: basePath should be in the repo, i.e. within {0!r}\n\n".format(repoPath))
            sys.stderr.write(usage)
            sys.exit(11)
    oldName = ""
    if oldRef:
        oldRev = repo.rev_parse(oldRef)
        oldName = oldRev.hexsha[:10]
        oldValues = loadRef(oldRev, relPath)
        sys.stdout.write("<old>: {relPath} in {oldRef}({oldName}) of repo at {repoPath}\n".format(
            relPath = relPath, oldRef = oldRef, oldName = oldName, repoPath = repoPath))
    else:
        sys.stdout.write("<old>: {oldPath}\n".format(oldPath = oldPath))
        oldValues = loadPath(oldPath)
    newName = ""
    if newRef:
        newRev = repo.rev_parse(newRef)
        newName = newRev.hexsha[:10]
        newValues = loadRef(newRev, relPath)
        sys.stdout.write("<new>: {relPath} in {newRef}({newName}) of repo at {repoPath}\n".format(
            relPath = relPath, newRef = newRef, newName = newName, repoPath = repoPath))
    else:
        sys.stdout.write("<new>: {newPath}\n".format(newPath = newPath))
        newValues = loadPath(newPath)
    res = buildOverrides(oldValues["envs"], newValues["envs"])
    if verbose:
        jsonIndentF(res, sys.stdout, check_circular = False)
        sys.stdout.write("\n")
    sys.stdout.flush()
    overrides = res["overrides"]
    if overrides and not checkOnly:
        overridesName = os.path.join(overridesDir, "{old}_{new}_{date}.nomadmetainfooverrides.json"
                .format(old = oldName, new = newName, date = datetime.date.today().isoformat()))
        n = 1
        while noClobber and os.path.exists(overridesName):
            overridesName = os.path.join(overridesDir, "{old}_{new}_{date}_{n}.nomadmetainfooverrides.json"
                    .format(old = oldName, new = newName, date = datetime.date.today().isoformat()), n = n)
            n += 1
        if os.path.exists(overridesName):
            overwriteStr = "Overwriting"
        else:
            overwriteStr = "Writing"
        with open(overridesName, "w") as f: # use os.open to guarantee no clashes seems like an overkill
            jsonIndentF(overrides, f, check_circular = False)
        sys.stdout.write("{0} overrides in {1!r}\n".format(overwriteStr, os.path.abspath(overridesName)))
    hasWarnings = oldValues["hasWarnings"] or newValues["hasWarnings"] or res["hasWarnings"]
    removals = res["removals"]
    if removals:
        sys.stdout.write("removals: ")
        jsonCompactF([x["name"] for x in removals], sys.stdout)
        sys.stdout.write("\n")
    additions = res["additions"]
    if additions:
        sys.stdout.write("additions: ")
        jsonCompactF([x["name"] for x in additions], sys.stdout)
        sys.stdout.write("\n")
    complexOverrides = res["complexOverrides"]
    if complexOverrides:
        hasWarnings = True
        sys.stdout.write("Warning: complexOverrides: ")
        jsonCompactF([x["name"] for x in complexOverrides], sys.stdout)
        sys.stdout.write("\n")
    sys.stdout.write("#overrides:{nov:3} #matches:{nm:3} #complexOverrides:{ncov:3} #additions:{nadd:3} #removals:{nrm}\n"
            .format(nm = res["nMatched"], nov = len(overrides), ncov = len(complexOverrides), nadd = len(additions), nrm = len(removals)))
    if not hasWarnings:
        sys.exit(0)
    else:
        sys.stderr("Had warnings\n")
        sys.exit(1)



import copy
from lark import Transformer, v_args, Token
from optimade.server.mappers import BaseResourceMapper


class MongoTransformer(Transformer):
    """Support for grammar v0.10.1"""

    operator_map = {
        "<": "$lt",
        "<=": "$lte",
        ">": "$gt",
        ">=": "$gte",
        "!=": "$ne",
        "=": "$eq",
    }
    reversed_operator_map = {
        "$lt": "$gt",
        "$lte": "$gte",
        "$gt": "$lt",
        "$gte": "$lte",
        "$ne": "$ne",
        "$eq": "$eq",
    }

    def __init__(self, mapper: BaseResourceMapper = None):
        """ Initialise the object, optionally loading in a
        resource mapper for use when post-processing.

        """
        self.mapper = mapper
        super().__init__()

    def postprocess(self, query):
        """ Used to post-process the final parsed query. """
        if self.mapper:
            # important to apply length alias before normal aliases
            query = self._apply_length_aliases(query)
            query = self._apply_aliases(query)

        query = self._apply_relationship_filtering(query)
        query = self._apply_length_operators(query)
        query = self._apply_unknown_or_null_filter(query)

        return query

    def transform(self, tree):
        return self.postprocess(super().transform(tree))

    def filter(self, arg):
        # filter: expression*
        return arg[0] if arg else None

    @v_args(inline=True)
    def constant(self, value):
        # constant: string | number
        # Note: Do nothing!
        return value

    @v_args(inline=True)
    def value(self, value):
        # value: string | number | property
        # Note: Do nothing!
        return value

    @v_args(inline=True)
    def non_string_value(self, value):
        """ non_string_value: number | property """
        # Note: Do nothing!
        return value

    @v_args(inline=True)
    def not_implemented_string(self, value):
        """ not_implemented_string: value

        Raise NotImplementedError.
        For further information, see Materials-Consortia/OPTIMADE issue 157:
        https://github.com/Materials-Consortia/OPTIMADE/issues/157
        """
        raise NotImplementedError("Comparing strings is not yet implemented.")

    def value_list(self, arg):
        # value_list: [ OPERATOR ] value ( "," [ OPERATOR ] value )*
        # NOTE: no support for optional OPERATOR, yet, so this takes the
        # parsed values and returns an error if that is being attempted
        for value in arg:
            if str(value) in self.operator_map.keys():
                raise NotImplementedError(
                    f"OPERATOR {value} inside value_list {arg} not implemented."
                )

        return arg

    def value_zip(self, arg):
        # value_zip: [ OPERATOR ] value ":" [ OPERATOR ] value (":" [ OPERATOR ] value)*
        raise NotImplementedError

    def value_zip_list(self, arg):
        # value_zip_list: value_zip ( "," value_zip )*
        raise NotImplementedError

    def expression(self, arg):
        # expression: expression_clause ( OR expression_clause )
        # expression with and without 'OR'
        return {"$or": arg} if len(arg) > 1 else arg[0]

    def expression_clause(self, arg):
        # expression_clause: expression_phrase ( AND expression_phrase )*
        # expression_clause with and without 'AND'
        return {"$and": arg} if len(arg) > 1 else arg[0]

    def expression_phrase(self, arg):
        # expression_phrase: [ NOT ] ( comparison | "(" expression ")" )
        return self._recursive_expression_phrase(arg)

    @v_args(inline=True)
    def comparison(self, value):
        # comparison: constant_first_comparison | property_first_comparison
        # Note: Do nothing!
        return value

    def property_first_comparison(self, arg):
        # property_first_comparison: property ( value_op_rhs | known_op_rhs | fuzzy_string_op_rhs | set_op_rhs |
        # set_zip_op_rhs | length_op_rhs )
        return {arg[0]: arg[1]}

    def constant_first_comparison(self, arg):
        # constant_first_comparison: constant OPERATOR ( non_string_value | not_implemented_string )
        return {arg[2]: {self.reversed_operator_map[self.operator_map[arg[1]]]: arg[0]}}

    @v_args(inline=True)
    def value_op_rhs(self, operator, value):
        # value_op_rhs: OPERATOR value
        return {self.operator_map[operator]: value}

    def known_op_rhs(self, arg):
        # known_op_rhs: IS ( KNOWN | UNKNOWN )
        # The OPTIMADE spec also required a type comparison with null, this must be post-processed
        # so here we use a special key "#known" which will get replaced in post-processing with the
        # expanded dict
        return {"#known": arg[1] == "KNOWN"}

    def fuzzy_string_op_rhs(self, arg):
        # fuzzy_string_op_rhs: CONTAINS value | STARTS [ WITH ] value | ENDS [ WITH ] value

        # The WITH keyword may be omitted.
        if isinstance(arg[1], Token) and arg[1].type == "WITH":
            pattern = arg[2]
        else:
            pattern = arg[1]

        # CONTAINS
        if arg[0] == "CONTAINS":
            regex = f"{pattern}"
        elif arg[0] == "STARTS":
            regex = f"^{pattern}"
        elif arg[0] == "ENDS":
            regex = f"{pattern}$"
        return {"$regex": regex}

    def set_op_rhs(self, arg):
        # set_op_rhs: HAS ( [ OPERATOR ] value | ALL value_list | ANY value_list | ONLY value_list )

        if len(arg) == 2:
            # only value without OPERATOR
            return {"$in": arg[1:]}

        if arg[1] == "ALL":
            return {"$all": arg[2]}

        if arg[1] == "ANY":
            return {"$in": arg[2]}

        if arg[1] == "ONLY":
            return {"$all": arg[2], "$size": len(arg[2])}

        # value with OPERATOR
        raise NotImplementedError(
            f"set_op_rhs not implemented for use with OPERATOR. Given: {arg}"
        )

    def length_op_rhs(self, arg):
        # length_op_rhs: LENGTH [ OPERATOR ] value
        if len(arg) == 2 or (len(arg) == 3 and arg[1] == "="):
            return {"$size": arg[-1]}

        elif arg[1] in self.operator_map and arg[1] != "!=":
            # create an invalid query that needs to be post-processed
            # e.g. {'$size': {'$gt': 2}}, which is not allowed by Mongo.
            return {"$size": {self.operator_map[arg[1]]: arg[-1]}}

        raise NotImplementedError(
            f"Operator {arg[1]} not implemented for LENGTH filter."
        )

    def set_zip_op_rhs(self, arg):
        # set_zip_op_rhs: property_zip_addon HAS ( value_zip | ONLY value_zip_list | ALL value_zip_list |
        # ANY value_zip_list )
        raise NotImplementedError

    def property_zip_addon(self, arg):
        # property_zip_addon: ":" property (":" property)*
        raise NotImplementedError

    def property(self, arg):
        # property: IDENTIFIER ( "." IDENTIFIER )*
        return ".".join(arg)

    @v_args(inline=True)
    def string(self, string):
        # string: ESCAPED_STRING
        return string.strip('"')

    @v_args(inline=True)
    def signed_int(self, number):
        # signed_int : SIGNED_INT
        return int(number)

    @v_args(inline=True)
    def number(self, number):
        # number: SIGNED_INT | SIGNED_FLOAT
        if number.type == "SIGNED_INT":
            type_ = int
        elif number.type == "SIGNED_FLOAT":
            type_ = float
        return type_(number)

    def __default__(self, data, children, meta):
        raise NotImplementedError(
            f"Calling __default__, i.e., unknown grammar concept. data: {data}, children: {children}, meta: {meta}"
        )

    def _recursive_expression_phrase(self, arg):
        """ Helper function for parsing `expression_phrase`. Recursively sorts out
        the correct precedence for $and, $or and $nor.

        """
        if len(arg) == 1:
            # without NOT
            return arg[0]

        # handle the case of {"$not": {"$or": [expr1, expr2]}} using {"$nor": [expr1, expr2]}.
        if "$or" in arg[1]:
            return {"$nor": self._recursive_expression_phrase([arg[1]["$or"]])}

        # handle the case of {"$not": {"$and": [expr1, expr2]}} using per-expression negation,
        # e.g. {"$and": [{prop1: {"$not": expr1}}, {prop2: {"$not": ~expr2}}]}.
        # Note that this is not the same as NOT (expr1 AND expr2)!
        if "$and" in arg[1]:
            return {
                "$and": [
                    self._recursive_expression_phrase(["NOT", subdict])
                    for subdict in arg[1]["$and"]
                ]
            }

        # simple case of negating one expression, from NOT (expr) to ~expr.
        return {prop: {"$not": expr} for prop, expr in arg[1].items()}

    def _apply_length_aliases(self, filter_: dict) -> dict:
        """ Recursively search query for any $size calls, and check
        if the property can be replaced with its corresponding length
        alias.

        """

        def check_for_size(prop, expr):
            return (
                isinstance(expr, dict)
                and "$size" in expr
                and self.mapper.length_alias_for(prop)
            )

        def replace_with_length_alias(subdict, prop, expr):
            subdict[self.mapper.length_alias_for(prop)] = expr["$size"]
            subdict[prop].pop("$size")
            if not subdict[prop]:
                subdict.pop(prop)
            return subdict

        return recursive_postprocessing(
            filter_, check_for_size, replace_with_length_alias
        )

    def _apply_aliases(self, filter_: dict) -> dict:
        """ Check whether any fields in the filter have aliases so
        that they can be renamed for the Mongo query.

        """
        # if there are no defined aliases, just skip
        if not self.mapper.all_aliases():
            return filter_

        def check_for_alias(prop, expr):
            return self.mapper.alias_for(prop) != prop

        def apply_alias(subdict, prop, expr):
            if isinstance(subdict, dict):
                subdict[self.mapper.alias_for(prop)] = self._apply_aliases(
                    subdict.pop(prop)
                )
            elif isinstance(subdict, str):
                subdict = self.mapper.alias_for(subdict)

            return subdict

        return recursive_postprocessing(filter_, check_for_alias, apply_alias)

    def _apply_length_operators(self, filter_: dict) -> dict:
        """ Check for any invalid pymongo queries that involve
        applying an operator to the length of a field, and transform
        them into a test for existence of the relevant entry, e.g.
        "list LENGTH > 3" becomes "does the 4th list entry exist?".

        """

        def check_for_length_op_filter(prop, expr):
            return (
                isinstance(expr, dict)
                and "$size" in expr
                and isinstance(expr["$size"], dict)
            )

        def apply_length_op(subdict, prop, expr):
            # assumes that the dictionary only has one element by design
            # (we just made it above in the transformer)
            operator, value = list(expr["$size"].items())[0]
            if operator in self.operator_map.values() and operator != "$ne":
                # worth being explicit here, I think
                _prop = None
                existence = None
                if operator == "$gt":
                    _prop = f"{prop}.{value + 1}"
                    existence = True
                elif operator == "$gte":
                    _prop = f"{prop}.{value}"
                    existence = True
                elif operator == "$lt":
                    _prop = f"{prop}.{value}"
                    existence = False
                elif operator == "$lte":
                    _prop = f"{prop}.{value + 1}"
                    existence = False
                if _prop is not None:
                    subdict.pop(prop)
                    subdict[_prop] = {"$exists": existence}

            return subdict

        return recursive_postprocessing(
            filter_, check_for_length_op_filter, apply_length_op,
        )

    def _apply_relationship_filtering(self, filter_: dict) -> dict:
        """ Check query for property names that match the entry
        types, and transform them as relationship filters rather than
        property filters.

        """

        def check_for_entry_type(prop, expr):
            return str(prop).count(".") == 1 and str(prop).split(".")[0] in (
                "structures",
                "references",
            )

        def replace_with_relationship(subdict, prop, expr):
            _prop, _field = str(prop).split(".")
            if _field != "id":
                raise NotImplementedError(
                    f'Cannot filter relationships by field "{_field}", only "id" is supported.'
                )

            # in the case of HAS ONLY, the size operator needs to be applied
            # one level up, i.e. excluding the field
            if "$size" in expr:
                if "$and" not in subdict:
                    subdict["$and"] = []
                subdict["$and"].extend(
                    [
                        {f"relationships.{_prop}.data": {"$size": expr.pop("$size")}},
                        {f"relationships.{_prop}.data.{_field}": expr},
                    ]
                )
            else:
                subdict[f"relationships.{_prop}.data.{_field}"] = expr

            subdict.pop(prop)

            return subdict

        return recursive_postprocessing(
            filter_, check_for_entry_type, replace_with_relationship
        )

    def _apply_unknown_or_null_filter(self, filter_: dict) -> dict:
        """ This method loops through the query and replaces the check for
        KNOWN with a check for existence and a check for not null, and the
        inverse for UNKNOWN.

        """

        def check_for_known_filter(_, expr):
            """ Find cases where the query dict looks like
            `{"field": {"#known": T/F}}` or
            `{"field": "$not": {"#known": T/F}}`, which is a magic word
            for KNOWN/UNKNOWN filters in this transformer.

            """
            return isinstance(expr, dict) and (
                "#known" in expr or "#known" in expr.get("$not", {})
            )

        def replace_known_filter_with_or(subdict, prop, expr):
            """ Replace #known and $not->#known parsed filters with the appropriate
            combination of $exists and/or $eq/$ne null.

            """
            not_ = set(expr.keys()) == {"$not"}
            if not_:
                expr = expr["$not"]

            exists = expr["#known"] ^ not_

            top_level_key = "$or"
            comparison_operator = "$eq"
            if exists:
                top_level_key = "$and"
                comparison_operator = "$ne"

            if top_level_key not in subdict:
                subdict[top_level_key] = []

            subdict[top_level_key].append({prop: {"$exists": exists}})
            subdict[top_level_key].append({prop: {comparison_operator: None}})

            subdict.pop(prop)

            return subdict

        return recursive_postprocessing(
            filter_, check_for_known_filter, replace_known_filter_with_or
        )


def recursive_postprocessing(filter_, condition, replacement):
    """ Recursively descend into the query, checking each dictionary
    (contained in a list, or as an entry in another dictionary) for
    the condition passed. If the condition is true, apply the
    replacement to the dictionary.

    Parameters:
        filter_ (list/dict): the filter_ to process.
        condition (callable): a function that returns True if the
            replacement function should be applied. It should take
            as arguments the property and expression from the filter_,
            as would be returned by iterating over `filter_.items()`.
        replacement (callable): a function that returns the processed
            dictionary. It should take as arguments the dictionary
            to modify, the property and the expression (as described
            above).

    Example:
        For the simple case of replacing one field name with
        another, the following functions could be used:

        ```python
        def condition(prop, expr):
            return prop == "field_name_old"

        def replacement(d, prop, expr):
            d["field_name_old"] = d.pop(prop)

        filter_ = recursive_postprocessing(
            filter_, condition, replacement
        )

        ```

    """
    if isinstance(filter_, list):
        result = [recursive_postprocessing(q, condition, replacement) for q in filter_]
        return result

    if isinstance(filter_, dict):
        # this could potentially lead to memory leaks if the filter_ is *heavily* nested
        _cached_filter = copy.deepcopy(filter_)
        for prop, expr in filter_.items():
            if condition(prop, expr):
                _cached_filter = replacement(_cached_filter, prop, expr)
            elif isinstance(expr, list):
                _cached_filter[prop] = [
                    recursive_postprocessing(q, condition, replacement) for q in expr
                ]
        return _cached_filter

    return filter_



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='octopus_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='octopus.nomadmetainfo.json'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_octopus_info_energy_ion_ion = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ion-ion interaction energy
        ''',
        a_legacy=LegacyDefinition(name='x_octopus_info_energy_ion_ion'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_octopus_info_scf_converged_iterations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of scf iterations to converge calculation
        ''',
        a_legacy=LegacyDefinition(name='x_octopus_info_scf_converged_iterations'))

    x_octopus_log_svn_revision = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        octopus svn revision
        ''',
        a_legacy=LegacyDefinition(name='x_octopus_log_svn_revision'))


m_package.__init_metainfo__()



from pathlib import Path

import pytest
import unittest
import json

from pydantic import ValidationError  # pylint: disable=no-name-in-module
from optimade.models import (
    StructureResource,
    EntryRelationships,
    ReferenceResource,
    AvailableApiVersion,
)
from optimade.models.jsonapi import Error
from optimade.server.mappers import StructureMapper, ReferenceMapper


class TestPydanticValidation(unittest.TestCase):
    def test_good_structures(self):
        test_structures_path = (
            Path(__file__)
            .resolve()
            .parent.joinpath("../../optimade/server/data/test_structures.json")
        )
        with open(test_structures_path, "r") as f:
            good_structures = json.load(f)
            # adjust dates as mongo would
        for doc in good_structures:
            doc["last_modified"] = doc["last_modified"]["$date"]

        for structure in good_structures:
            StructureResource(**StructureMapper.map_back(structure))

    def test_more_good_structures(self):
        test_structures_path = (
            Path(__file__).resolve().parent.joinpath("test_more_structures.json")
        )
        with open(test_structures_path, "r") as f:
            good_structures = json.load(f)

        # adjust dates as mongo would
        for doc in good_structures:
            doc["last_modified"] = doc["last_modified"]["$date"]

        for structure in good_structures:
            StructureResource(**StructureMapper.map_back(structure))

    def test_bad_structures(self):
        test_structures_path = (
            Path(__file__).resolve().parent.joinpath("test_bad_structures.json")
        )
        with open(test_structures_path, "r") as f:
            bad_structures = json.load(f)
        for doc in bad_structures:
            doc["last_modified"] = doc["last_modified"]["$date"]

        for ind, structure in enumerate(bad_structures):
            with self.assertRaises(
                ValidationError,
                msg="Bad test structure {} failed to raise an error\nContents: {}".format(
                    ind, json.dumps(structure, indent=2)
                ),
            ):
                StructureResource(**StructureMapper.map_back(structure))

    def test_simple_relationships(self):
        """Make sure relationship resources are added to the correct relationship"""

        good_relationships = (
            {"references": {"data": [{"id": "dijkstra1968", "type": "references"}]}},
            {"structures": {"data": [{"id": "dijkstra1968", "type": "structures"}]}},
        )
        for relationship in good_relationships:
            EntryRelationships(**relationship)

        bad_relationships = (
            {"references": {"data": [{"id": "dijkstra1968", "type": "structures"}]}},
            {"structures": {"data": [{"id": "dijkstra1968", "type": "references"}]}},
        )
        for relationship in bad_relationships:
            with self.assertRaises(ValidationError):
                EntryRelationships(**relationship)

    def test_advanced_relationships(self):
        """Make sure the rules for the base resource 'meta' field are upheld"""

        relationship = {
            "references": {
                "data": [
                    {
                        "id": "dijkstra1968",
                        "type": "references",
                        "meta": {
                            "description": "Reference for the search algorithm Dijkstra."
                        },
                    }
                ]
            }
        }
        EntryRelationships(**relationship)

        relationship = {
            "references": {
                "data": [{"id": "dijkstra1968", "type": "references", "meta": {}}]
            }
        }
        with self.assertRaises(ValidationError):
            EntryRelationships(**relationship)

    def test_good_references(self):
        test_refs_path = (
            Path(__file__)
            .resolve()
            .parent.joinpath("../../optimade/server/data/test_references.json")
        )
        with open(test_refs_path, "r") as f:
            good_refs = json.load(f)
        for doc in good_refs:
            doc["last_modified"] = doc["last_modified"]["$date"]
            ReferenceResource(**ReferenceMapper.map_back(doc))

    def test_bad_references(self):
        bad_refs = [
            {"id": "AAAA", "type": "references", "doi": "10.1234/1234"},  # bad id
            {"id": "newton1687", "type": "references"},  # missing all fields
            {
                "id": "newton1687",
                "type": "reference",
                "doi": "10.1234/1234",
            },  # wrong type
        ]

        for ref in bad_refs:
            with self.assertRaises(ValidationError):
                ReferenceResource(**ReferenceMapper.map_back(ref))

    def test_available_api_versions(self):
        bad_urls = [
            {"url": "asfdsafhttps://example.com/v0.0", "version": "0.0.0"},
            {"url": "https://example.com/optimade", "version": "1.0.0"},
            {"url": "https://example.com/v0999", "version": "0999.0.0"},
            {"url": "http://example.com/v2.3", "version": "2.3.0"},
        ]
        good_urls = [
            {"url": "https://example.com/v0", "version": "0.1.9"},
            {"url": "https://example.com/v1.0.2", "version": "1.0.2"},
            {"url": "https://example.com/optimade/v1.2", "version": "1.2.3"},
        ]
        bad_combos = [
            {"url": "https://example.com/v0", "version": "1.0.0"},
            {"url": "https://example.com/v1.0.2", "version": "1.0.3"},
            {"url": "https://example.com/optimade/v1.2", "version": "1.3.2"},
        ]

        for data in bad_urls:
            with pytest.raises(ValueError) as exc:
                AvailableApiVersion(**data)
                pytest.fail(f"Url {data['url']} should have failed")
            self.assertTrue(
                "url MUST be a versioned base URL" in exc.exconly()
                or "URL scheme not permitted" in exc.exconly(),
                msg=f"Validator 'url_must_be_versioned_base_url' not triggered as it should.\nException message: {exc.exconly()}.\nInputs: {data}",
            )

        for data in bad_combos:
            with pytest.raises(ValueError) as exc:
                AvailableApiVersion(**data)
                pytest.fail(
                    f"{data['url']} should have failed with version {data['version']}"
                )
            self.assertTrue(
                "is not compatible with url version" in exc.exconly(),
                msg=f"Validator 'crosscheck_url_and_version' not triggered as it should.\nException message: {exc.exconly()}.\nInputs: {data}",
            )

        for data in good_urls:
            self.assertIsInstance(AvailableApiVersion(**data), AvailableApiVersion)

    def test_hashability(self):
        error = Error(id="test")
        set([error])



import urllib

from typing import Union

from fastapi import APIRouter, Request
from fastapi.exceptions import StarletteHTTPException

from optimade import __api_version__

from optimade.models import (
    ErrorResponse,
    InfoResponse,
    EntryInfoResponse,
    ReferenceResource,
    StructureResource,
)

from .utils import meta_values, retrieve_queryable_properties, get_base_url


router = APIRouter(redirect_slashes=True)

ENTRY_INFO_SCHEMAS = {
    "structures": StructureResource.schema,
    "references": ReferenceResource.schema,
}


@router.get(
    "/info",
    response_model=Union[InfoResponse, ErrorResponse],
    response_model_exclude_unset=True,
    tags=["Info"],
)
def get_info(request: Request):
    from optimade.models import BaseInfoResource, BaseInfoAttributes

    parse_result = urllib.parse.urlparse(str(request.url))
    base_url = get_base_url(parse_result)

    return InfoResponse(
        meta=meta_values(str(request.url), 1, 1, more_data_available=False),
        data=BaseInfoResource(
            id=BaseInfoResource.schema()["properties"]["id"]["const"],
            type=BaseInfoResource.schema()["properties"]["type"]["const"],
            attributes=BaseInfoAttributes(
                api_version=f"v{__api_version__}",
                available_api_versions=[
                    {
                        "url": f"{base_url}/v{__api_version__.split('.')[0]}",
                        "version": __api_version__,
                    }
                ],
                formats=["json"],
                available_endpoints=["info", "links"] + list(ENTRY_INFO_SCHEMAS.keys()),
                entry_types_by_format={"json": list(ENTRY_INFO_SCHEMAS.keys())},
                is_index=False,
            ),
        ),
    )


@router.get(
    "/info/{entry}",
    response_model=Union[EntryInfoResponse, ErrorResponse],
    response_model_exclude_unset=True,
    tags=["Info"],
)
def get_entry_info(request: Request, entry: str):
    from optimade.models import EntryInfoResource

    valid_entry_info_endpoints = ENTRY_INFO_SCHEMAS.keys()
    if entry not in valid_entry_info_endpoints:
        raise StarletteHTTPException(
            status_code=404,
            detail=f"Entry info not found for {entry}, valid entry info endpoints are: {valid_entry_info_endpoints}",
        )

    schema = ENTRY_INFO_SCHEMAS[entry]()
    queryable_properties = {"id", "type", "attributes"}
    properties = retrieve_queryable_properties(schema, queryable_properties)

    output_fields_by_format = {"json": list(properties.keys())}

    return EntryInfoResponse(
        meta=meta_values(str(request.url), 1, 1, more_data_available=False),
        data=EntryInfoResource(
            formats=list(output_fields_by_format.keys()),
            description=schema.get("description", "Entry Resources"),
            properties=properties,
            output_fields_by_format=output_fields_by_format,
        ),
    )



# Copyright 2016-2018 Mikkel Strange, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import division
import os
from contextlib import contextmanager
import numpy as np
from nomadcore.unit_conversion.unit_conversion import convert_unit as cu
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from .reader import Reader
import logging


@contextmanager
def open_section(p, name):
    gid = p.openSection(name)
    yield gid
    p.closeSection(name, gid)


def c(value, unit=None):
    """ Dummy function for unit conversion"""
    return cu(value, unit)


parser_info = {"name": "parser_mopac", "version": "1.0"}

def parse(filename, backend):
    p = backend
    o = open_section
    r = Reader(filename)
    p.startedParsingSession(filename, parser_info)

    with o(p, 'section_run'):
        p.addValue('program_name', 'MOPAC')
        p.addValue('program_version', r.program_version)
        with o(p, 'section_system') as system_gid:
            p.addArrayValues('atom_labels', r.atom_labels)
            p.addArrayValues('atom_positions', c(r.atom_positions,
                                                 'angstrom'))
            pbc = np.array((0, 0, 0), bool)
            p.addArrayValues('configuration_periodic_dimensions', pbc)
            if 'number_of_electrons' in r:
                p.addArrayValues('number_of_electrons', r.number_of_electrons)
        with o(p, 'section_sampling_method'):
            p.addValue('ensemble_type', 'NVE')
        with o(p, 'section_frame_sequence'):
            pass
        with o(p, 'section_method') as method_gid:
            p.addValue('x_mopac_method', r.x_mopac_method)
            p.addValue('x_mopac_keyword_line', r.inp_parm_line)
            p.addRealValue('total_charge', r.total_charge)
#            p.addValue('scf_threshold_energy_change',
#                       c(r.scf_threshold_energy_changer, 'hartree'))
            if 'spin_target_multiplicity' in r:
                p.addRealValue('spin_target_multiplicity',
                               r.spin_target_multiplicity)
        with o(p, 'section_single_configuration_calculation'):
            p.addValue('single_configuration_calculation_to_system_ref',
                       system_gid)
            p.addValue('single_configuration_to_calculation_method_ref',
                       method_gid)
            p.addRealValue('energy_total', c(r.energy_total, 'eV'))
            if 'x_mopac_fhof' in r:
                p.addRealValue('x_mopac_fhof', c(r.x_mopac_fhof, 'eV'))
            if 'electronic_kinetic_energy' in r:
                p.addRealValue('electronic_kinetic_energy',
                               c(r.electronic_kinetic_energy, 'eV'))
            if 'atom_forces' in r:
                p.addArrayValues('atom_forces',
                                 c(r.atom_forces, 'eV/angstrom'))
            if 'energy_reference_highest_occupied' in r:
                p.addArrayValues('energy_reference_highest_occupied',
                               c(r.energy_reference_highest_occupied, 'eV'))
            if 'spin_S2' in r:
                p.addRealValue('spin_S2', r.spin_S2)

            if 'time_calculation' in r:
                p.addRealValue('time_calculation', r.time_calculation)

            with o(p, 'section_eigenvalues'):
                if 'eigenvalues_values' in r:
                    p.addValue('eigenvalues_kind', 'normal')
                    p.addArrayValues('eigenvalues_values',
                                     c(r.eigenvalues_values, 'eV'))
                    p.addArrayValues('eigenvalues_kpoints',
                                     r.eigenvalues_kpoints)
                if 'eigenvalues_occupation' in r:
                    p.addArrayValues('eigenvalues_occupation',
                                     r.eigenvalues_occupation)

    p.finishedParsingSession("ParseSuccess", None)


class MopacParser():
    """ A proper class envolop for running this parser from within python. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
        logging.info('mopac parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("mopac.nomadmetainfo.json")
        parse(mainfile, backend)
        return backend


if __name__ == '__main__':
    import sys
    filename = sys.argv[1]
    parse(filename)



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os

from nomadcore.local_meta_info import loadJsonFile, InfoKindEl

############################################################
# This file contains functions that are needed
# by more than one parser.
############################################################

# loading metadata from nomad-meta-info/meta_info/nomad_meta_info/lammps.nomadmetainfo.json
metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../../../nomad-meta-info/meta_info/nomad_meta_info/lammps.nomadmetainfo.json"))


def get_metaInfo(filePath=metaInfoPath):
    """Loads metadata.

    Args:
        filePath: Location of metadata.

    Returns:
        metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
    """
    metaInfoEnv, warnings = loadJsonFile(
        filePath=filePath,
        dependencyLoader=None,
        extraArgsHandling=InfoKindEl.ADD_EXTRA_ARGS,
        uri=None)

    return metaInfoEnv


#  List of unit conversion coefficient for all LAMMPS units styles
#  Multiply relevant parsed values by "toProperty" to convert to NOMAD units
class converter(object):

    def __init__(self, unitsType = 'lj'):
        self.name = unitsType

        if unitsType == 'lj':
            # mass = mass or m
            # distance = sigma, where x* = x / sigma
            # time = tau, where t* = t (epsilon / m / sigma^2)^1/2
            # energy = epsilon, where E* = E / epsilon
            # velocity = sigma/tau, where v* = v tau / sigma
            # force = epsilon/sigma, where f* = f sigma / epsilon
            # torque = epsilon, where t* = t / epsilon
            # temperature = reduced LJ temperature, where T* = T Kb / epsilon
            # pressure = reduced LJ pressure, where P* = P sigma^3 / epsilon
            # dynamic viscosity = reduced LJ viscosity, where eta* = eta sigma^3 / epsilon / tau
            # charge = reduced LJ charge, where q* = q / (4 pi perm0 sigma epsilon)^1/2
            # dipole = reduced LJ dipole, moment where *mu = mu / (4 pi perm0 sigma^3 epsilon)^1/2
            # electric field = force/charge, where E* = E (4 pi perm0 sigma epsilon)^1/2 sigma / epsilon
            # density = mass/volume, where rho* = rho sigma^dim
            ########################################################################################################
            # Conversion not supported!
            ########################################################################################################
            self.ratioMass     = 1
            self.ratioDistance = 1
            self.ratioTime     = 1
            self.ratioEnergy   = 1
            self.ratioVelocity = 1
            self.ratioForce    = 1
            self.ratioTorque   = 1
            self.ratioTemp     = 1
            self.ratioPress    = 1
            self.ratioDVisc    = 1
            self.ratioCharge   = 1
            self.ratioDipole   = 1
            self.ratioEField   = 1
            self.ratioDensity  = 1

        if unitsType == 'real':

            # Define unit conversion parameters for "units = real"
            # For style real, these are the default LAMMPS units:
            #
            #     mass = grams/mole
            #     distance = Angstroms
            #     time = femtoseconds
            #     energy = Kcal/mole
            #     velocity = Angstroms/femtosecond
            #     force = Kcal/mole-Angstrom
            #     torque = Kcal/mole
            #     temperature = Kelvin
            #     pressure = atmospheres
            #     dynamic viscosity = Poise
            #     charge = multiple of electron charge (1.0 is a proton)
            #     dipole = charge*Angstroms
            #     electric field = volts/Angstrom
            #     density = gram/cm^dim

            self.ratioMass     = 1.66054e-27
            self.ratioDistance = 1e-10
            self.ratioTime     = 1e-15
            self.ratioEnergy   = 6.94786e-21
            self.ratioVelocity = 1e5
            self.ratioForce    = 6.94786e-11
            self.ratioTorque   = 6.94786e-21
            self.ratioTemp     = 1
            self.ratioPress    = 1.01325e5
            self.ratioDVisc    = 0
            self.ratioCharge   = 1.602176565e-19
            self.ratioDipole   = 0
            self.ratioEField   = 0
            self.ratioDensity  = 0



        if unitsType == 'metal':

            # Define unit conversion parameters for "units = metal"
            # For style metal, these are the default LAMMPS units:
            #
            #   mass = grams/mole
            #   distance = Angstroms
            #   time = picoseconds
            #   energy = eV
            #   velocity = Angstroms/picosecond
            #   force = eV/Angstrom
            #   torque = eV
            #   temperature = Kelvin
            #   pressure = bars
            #   dynamic viscosity = Poise
            #   charge = multiple of electron charge (1.0 is a proton)
            #   dipole = charge*Angstroms
            #   electric field = volts/Angstrom
            #   density = gram/cm^dim

            self.ratioMass     = 1.66054e-27
            self.ratioDistance = 1e-10
            self.ratioTime     = 1e-12
            self.ratioEnergy   = 1.60218e-19
            self.ratioVelocity = 1e2
            self.ratioForce    = 1.60218e-9
            self.ratioTorque   = 1.60218e-19
            self.ratioTemp     = 1
            self.ratioPress    = 1e5
            self.ratioDVisc    = 0
            self.ratioCharge   = 1.602176565e-19
            self.ratioDipole   = 0
            self.ratioEField   = 0
            self.ratioDensity  = 0



        if unitsType == 'si':

            # Define unit conversion parameters for "units = si"
            # For style si, these are the default LAMMPS units:
            #
            #   mass = kilograms
            #   distance = meters
            #   time = seconds
            #   energy = Joules
            #   velocity = meters/second
            #   force = Newtons
            #   torque = Newton-meters
            #   temperature = Kelvin
            #   pressure = Pascals
            #   dynamic viscosity = Pascal*second
            #   charge = Coulombs (1.6021765e-19 is a proton)
            #   dipole = Coulombs*meters
            #   electric field = volts/meter
            #   density = kilograms/meter^dim

            self.ratioMass     = 1
            self.ratioDistance = 1
            self.ratioTime     = 1
            self.ratioEnergy   = 1
            self.ratioVelocity = 1
            self.ratioForce    = 1
            self.ratioTorque   = 1
            self.ratioTemp     = 1
            self.ratioPress    = 1
            self.ratioDVisc    = 1
            self.ratioCharge   = 1
            self.ratioDipole   = 1
            self.ratioEField   = 1
            self.ratioDensity  = 1


        if unitsType == 'cgs':

            # Define unit conversion parameters for "units = cgs"
            # For style cgs, these are the default LAMMPS units:
            #
            #   mass = grams
            #   distance = centimeters
            #   time = seconds
            #   energy = ergs
            #   velocity = centimeters/second
            #   force = dynes
            #   torque = dyne-centimeters
            #   temperature = Kelvin
            #   pressure = dyne/cm^2 or barye = 1.0e-6 bars
            #   dynamic viscosity = Poise
            #   charge = statcoulombs or esu (4.8032044e-10 is a proton)
            #   dipole = statcoul-cm = 10^18 debye
            #   electric field = statvolt/cm or dyne/esu
            #   density = grams/cm^dim

            self.ratioMass     = 1e-3
            self.ratioDistance = 1e-2
            self.ratioTime     = 1
            self.ratioEnergy   = 1e-7
            self.ratioVelocity = 1e-2
            self.ratioForce    = 1e-5
            self.ratioTorque   = 1e-7
            self.ratioTemp     = 1
            self.ratioPress    = 1e-1
            self.ratioDVisc    = 1
            self.ratioCharge   = 3.335640951982e-10
            self.ratioDipole   = 0
            self.ratioEField   = 0
            self.ratioDensity  = 0

        if unitsType == 'electron':

            # Define unit conversion parameters for "units = electron"
            # For style electron, these are the default LAMMPS units:
            #
            # mass = atomic mass units
            # distance = Bohr
            # time = femtoseconds
            # energy = Hartrees
            # velocity = Bohr/atomic time units [1.03275e-15 seconds]
            # force = Hartrees/Bohr
            # temperature = Kelvin
            # pressure = Pascals
            # charge = multiple of electron charge (1.0 is a proton)
            # dipole moment = Debye
            # electric field = volts/cm

            self.ratioMass     = 1.66054e-27
            self.ratioDistance = 5.29177249e-11
            self.ratioTime     = 1e-15
            self.ratioEnergy   = 4.35974e-18
            self.ratioVelocity = 1e-2
            self.ratioForce    = 8.2387264e-8
            self.ratioTorque   = 4.35974e-18
            self.ratioTemp     = 1
            self.ratioPress    = 1
            self.ratioDVisc    = 0
            self.ratioCharge   = 1.602176565e-19
            self.ratioDipole   = 0
            self.ratioEField   = 0
            self.ratioDensity  = 0

        if unitsType == 'micro':

            # Define unit conversion parameters for "units = micro"
            # For style micro, these are the default LAMMPS units:
            #
            #   mass = picograms
            #   distance = micrometers
            #   time = microseconds
            #   energy = picogram-micrometer^2/microsecond^2
            #   velocity = micrometers/microsecond
            #   force = picogram-micrometer/microsecond^2
            #   torque = picogram-micrometer^2/microsecond^2
            #   temperature = Kelvin
            #   pressure = picogram/(micrometer-microsecond^2)
            #   dynamic viscosity = picogram/(micrometer-microsecond)
            #   charge = picocoulombs (1.6021765e-7 is a proton)
            #   dipole = picocoulomb-micrometer
            #   electric field = volt/micrometer
            #   density = picograms/micrometer^dim

            self.ratioMass     = 1e-15
            self.ratioDistance = 1e-6
            self.ratioTime     = 1e-6
            self.ratioEnergy   = 1e-15
            self.ratconveioVelocity = 1
            self.ratioForce    = 1e-9
            self.ratioTorque   = 1e-15
            self.ratioTemp     = 1
            self.ratioPress    = 1e3
            self.ratioDVisc    = 0
            self.ratioCharge   = 1e-12
            self.ratioDipole   = 0
            self.ratioEField   = 0
            self.ratioDensity  = 0

        if unitsType == 'nano':

            # Define unit conversion parameters for "units = nano"
            # For style nano, these are the default LAMMPS units:
            #
            #   mass = attograms
            #   distance = nanometers
            #   time = nanoseconds
            #   energy = attogram-nanometer^2/nanosecond^2
            #   velocity = nanometers/nanosecond
            #   force = attogram-nanometer/nanosecond^2
            #   torque = attogram-nanometer^2/nanosecond^2
            #   temperature = Kelvin
            #   pressure = attogram/(nanometer-nanosecond^2)
            #   dynamic viscosity = attogram/(nanometer-nanosecond)
            #   charge = multiple of electron charge (1.0 is a proton)
            #   dipole = charge-nanometer
            #   electric field = volt/nanometer
            #   density = attograms/nanometer^dim

            self.ratioMass     = 1e-21
            self.ratioDistance = 1e-9
            self.ratioTime     = 1e-9
            self.ratioEnergy   = 1e-21
            self.ratioVelocity = 1
            self.ratioForce    = 1e-12
            self.ratioTorque   = 1e-21
            self.ratioTemp     = 1
            self.ratioPress    = 1e6
            self.ratioDVisc    = 0
            self.ratioCharge   = 1.602176565e-19
            self.ratioDipole   = 0
            self.ratioEField   = 0
            self.ratioDensity  = 0

    def Mass(self, val):
        return self.ratioMass * val

    def Distance(self, val):
        return self.ratioDistance * val

    def Time(self, val):
        return self.ratioTime * val

    def Energy(self, val):
        return self.ratioEnergy * val

    def Velocity(self, val):
        return self.ratioVelocity * val

    def Force(self, val):
        return self.ratioForce * val

    def Torque(self, val):
        return self.ratioTorque * val

    def Temp(self, val):
        return self.ratioTemp * val

    def Press(self, val):
        return self.ratioPress * val

    def DVisc(self, val):
        return self.ratioDVisc * val

    def Charge(self, val):
        return self.ratioCharge * val

    def Dipole(self, val):
        return self.ratioDipole * val

    def EField(self, val):
        return self.ratioEField * val

    def Density(self, val):
        return self.ratioDensity * val



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import LAMMPS.AtomStyle as Atom
from LAMMPS.AtomStyle import atom_style
from LAMMPS import Units
from LAMMPS.BaseClasses import Lammps, AtomStyle


class LAMMPS(Lammps):

    def __init__(self):
        self.atom_style = Atom.Atomic()
        self.units      = Units.lj()

    @property
    def atom_style(self):
        return self.__atom_style

    @atom_style.setter
    def atom_style(self, atom_style):
        if isinstance(atom_style, AtomStyle):
            self.__atom_style = atom_style
        else:
            print("wrong")


    def __str__(self):

        ret_str = '\n'.join([
            self._row_format.format('Program name:', 'LAMMPS'),
            str(self.atom_style),
            str(self.units)
        ])

        return ret_str


if __name__ == '__main__':
    l = LAMMPS()
    l.atom_style = atom_style('bond')

    print(l)


# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import logging
import os

from nomad import config as nomad_config, utils


class LazyCommand(click.Command):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.orig_callback = self.callback
        self.callback = self.lazy_callback

    def lazy_callback(self, *args, **kwargs):
        for group_callback, group_args, group_kwargs in self.ctx.obj.group_invocations:
            group_callback(*group_args, **group_kwargs)
        return self.orig_callback(*args, **kwargs)

    def invoke(self, ctx):
        self.ctx = ctx
        return super().invoke(ctx)


class LazyGroup(click.Group):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.orig_callback = self.callback

        self.callback = self.lazy_callback
        self.ctx = None

    def lazy_callback(self, *args, **kwargs):
        self.ctx.obj.group_invocations.append((self.orig_callback, args, kwargs))
        return None

    def command(self, *args, **kwargs):
        kwargs.setdefault('cls', LazyCommand)
        return super().command(*args, **kwargs)

    def group(self, *args, **kwargs):
        kwargs.setdefault('cls', LazyGroup)
        return super().group(*args, **kwargs)

    def invoke(self, ctx):
        if ctx.obj is None:
            ctx.obj = POPO()
            ctx.obj.group_invocations = []

        self.ctx = ctx
        return super().invoke(ctx)


class POPO(dict):
    '''
    A dict subclass that uses attributes as key/value pairs.
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)


@click.group(
    cls=LazyGroup,
    help='''This is the entry point to nomad\'s command line interface CLI.
                     It uses a sub-command structure similar to the git command.''')
@click.option('-v', '--verbose', help='sets log level to info', is_flag=True)
@click.option('--debug', help='sets log level to debug', is_flag=True)
@click.option('--config', help='the config file to use')
@click.pass_context
def cli(ctx, verbose: bool, debug: bool, config: str):
    nomad_config.meta.service = os.environ.get('NOMAD_SERVICE', 'cli')

    if config is not None:
        nomad_config.load_config(config_file=config)

    if debug:
        utils.set_console_log_level(logging.DEBUG)
    elif verbose:
        utils.set_console_log_level(logging.INFO)
    else:
        utils.set_console_log_level(logging.WARNING)


def run_cli():
    try:
        return cli()  # pylint: disable=E1120,E1123
    except ImportError:
        import sys

        if next(arg for arg in sys.argv if arg == '-v') is not None:
            import traceback
            traceback.print_exc()

        print(
            'You are accessing functionality that requires extra dependencies.\n'
            'Check the NOMAD documentation or install all extra dependencies:\n'
            '  pip install nomad[all]', file=sys.stderr)

        sys.exit(1)
